/*---------------------------------------------------------------------------
**
** %GridDistribute.sas                                  01May2009
**
** ---------------------------------------------------------------------------
**                           DISCLAIMER
**
**   THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
**   TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
**   EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
**   PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
**   CONTAINED HEREIN.
** --------------------------------------------------------------------------
**
** The macros defined in this file are centered around the %Distribute
** macro, which uses MP CONNECT tools from SAS/CONNECT software to employ
** distributed processing on a grid for problems that consist of many
** replicate runs of a fundamental task.  For example, global integer
** optimization problems are often solved by steepest ascent (the fundamental
** task) with random restarts (the replicate runs).  Similarly, Monte Carlo
** methods in statistics rely on a set of pseudo-random observations (the
** replicate runs) of one or more difficult-to-compute test statistics
** (the fundamental task).  If your problem can be factored in this way,
** then %Distribute may offer a useful approach.
**
** Besides running the fundamental task you specify the required
** replicate number of times, %Distribute offers two additional
** services:
**
**    - Problems of the type that %Distribute can handle often require
**      a stream of random numbers, and %Distribute creates and
**      manages such a stream for each grid node.
**
**    - In order to monitor your program, you would like to know
**      where your fundamental processing time is spent.  So
**      %Distribute keeps track of how much time each grid node spends
**      working on its fundamental tasks and waiting for the client
**      to give it more work.
**
** ---------------------------------------------------
**  Creating the code
** ---------------------------------------------------
**
**  In order to use %Distribute system, you must first create the
**  %RInit macro which defines the programs to be run on each grid node.
**  The %RInit macro really defines the distributed computation.  It
**  needs to wrap of (at least) two macro definitions
**
**     %FirstRSub - initialization on that grid node
**     %TaskRSub  - the fundamental unit of distributed computation
**
**     within an RSUBMIT block, like so
**
**        %macro RInit;
**           rsubmit remote=Host&iHost wait=yes macvar=Status&iHost;
**
**              %macro FirstRSub;
**                 << Initialization for this grid node >>;
**              %mend;
**
**              %macro TaskRSub;
**                 << Fundamental unit of distributed computation >>;
**              %mend;
**
**              << Optional additional initializations >>;
**
**           endrsubmit;
**        %mend;
**
**  It is your job to write the macros %FirstRSub and %TaskRSub so
**  that they define the fundamental task.  As parameters for these
**  macros, you can assume the following global macro variables to
**  be available within the grid node SAS session:
**
**     Predefined
**     global macro      Contents
**     --------------    ----------------------------------------
**     Rem_Host          Name of this host
**     Rem_iHost         Index of this host out of all hosts
**     Rem_Seed          Random seed, rerandomized for each chunk
**     Rem_NIterAll      Total number of iterations
**     Rem_NIter         Number of Iterations in a chunk
**     Rem_JobIters      Number of Iterations in this chunk *
**     Rem_JobID         Chunk number *
**     GlobalNSub        Number of iterations already submitted *
**
**                       * = Available only in TaskRSub
**
**  The Rem_JobIters can be different from the Rem_NIter if you are
**  processing the last job and the number of iterations is not an
**  exact multiple of the number of iterations in a chunk. For example,
**  if the total iterations was 10 and the number of iterations in a
**  chunk was 3, the last chunk (chunk #4) would only have 1 iteration
**  in it (Chunk #1=3, Chunk#2=3, Chunk#3=3, Chunk #4=1)
**
**                         ---------
**  If you need other macro values (say, &Mac1, &Mac2, and &Mac3) to
**  be passed from the client to the grid node, you should specify
**  them in a %Macros macro, like so:
**
**     %macro Macros;
**        %syslput rem_Mac1=&Mac1;
**        %syslput rem_Mac2=&Mac2;
**        %syslput rem_Mac3=&Mac3;
**     %mend;
**
**  The %Distribute macro will take care of running this macro for
**  each grid node, so that you can use the macros &rem_Mac1,
**  &rem_Mac2, and &rem_Mac3 in your definitions of %FirstRSub and
**  %TaskRSub.
**
**
**                         ---------
**  If you need other things submitted to the grid node before each
**  TaskRSub is called, you can define a ClientRSub macro.
**
**     %macro ClientRSub;
**         %syslput rem_Mac4=&Mac4;
**         rsubmit remote=Host&iHost wait=yes macvar=Status&iHost;
**
**             << Optional additional initializations >>;
**
**         endrsubmit;
**     %mend;
**
**  The %Distribute macro will take care of running this macro on
**  the client before the RSUBMIT is made to run the TaskRSub.
**
** ---------------------------------------------------
**  Running your code
** ---------------------------------------------------
**
**  Once you have created the %RInit macro, you can distribute the
**  fundamental task across all the grid nodes with just the following
**  lines of code:
**
**     %let GridAppServer=<Name of SAS Application Server>;
**     %SignOn;
**     %let NIterAll = <Size of entire problem>;
**     %let NIter    = <Size of each chunk>;
**     %Distribute;
**     %SignOff;
**
**  Since accessing a SAS Application Server's Logical Grid Server
**  definition requires access to a metadata server, the program will
**  prompt you for the connection information to the metadata server.
**  If you do not want to be prompted, you can add metadata server
**  options similar to the following:
**
**     options metaserver="mymetadataserver.mycompany.com";
**     options metaport=8561;
**     options metarepository="Foundation";
**     options metauser="myusername";
**     options metapass="mypassword";
**
**
**  As chunks of the fundamental task are parceled out to a grid node, a
**  status line will be printed to the SAS log on the client, displaying
**     '?' for grid nodes that have not been contacted
**     '.' for grid nodes that are currently working
**     '!' for grid nodes that are currently waiting to be assigned work
**     '-' for grid nodes that have left the grid (via DIST_AUTOSIGNOFF)
**     'X' for grid nodes that are unable to do work for some reason.
**  Additional status information includes the number of fundamental tasks
**  submitted and completed so far.  When the distribution is complete,
**  the system will display a breakdown of how much work was done on each
**  grid node, together with an over-all summary of the efficiency of the
**  distribution.
**
** ---------------------------------------------------
**  Retrieving your results
** ---------------------------------------------------
**
**  There are a number of ways to retrieve the results of your code, namely,
**  proc upload, proc download, or Remote Library Services. Additionally,
**  you can use the %RCollect macro to collect the results from each
**  grid node into a data set or DATA step view that catenates all of
**  them. When each host responds to a signon, it is setup as a
**  Remote Library Services server. If the host is still connected at
**  the end of the grid run, you can retrieve all similarly named
**  datasets from each host.
**
**  NOTE: If more than one chunk is run on a host care must be taken to
**  append data to an existing data set instead of creating a new data
**  set on each chunk.
**
**  NOTE: Do not set DIST_AUTOSIGNOFF to anything other than 0 because
**  the grid node needs to be connected to use the RCollect macro.
**
**
**  Further analysis of the collected results can be performed on the
**  client.
**
** ---------------------------------------------------
**  Macro Variable Definitions:
** ---------------------------------------------------
**
**  NIterAll           Number of total iterations required. REQUIRED
**
**  NIter              Number of iterations per run. The total number
**                     of iterations divided by the number of iterations
**                     per run will determine the number of jobs that
**                     will be spawned. REQUIRED
**
**  GridAppServer      SAS Application Server name to use to get grid
**                     information. If not present or invalid, remote
**                     sessions will be spawned to the local machine based
**                     on the CPUCOUNT. The local sessions will be spawned
**                     using the sascmd option if present.
**
**  MaxGridNodesToUse  Maximum number of grid nodes to use for processing.
**                     Defaults to all available grid nodes.
**
**  MaxRunTime         Maximum amount of minutes allowed for all jobs to be
**                     distributed. Chunks will not be distributed after
**                     the max run minutes has expired.
**
**  DistLog            Log file used to store hidden GridDistribute source.
**  DistLst            Lst file used to store hidden GridDistribute listings.
**
**  DIST_STATUS        Determines if the status line is printed periodically.
**                     Defaults to 1 (Yes)
**
**  DIST_DETAIL        Determines if run detail is summarized at the end.
**                     Defaults to 1 (Yes)
**
**  DIST_CHART         Determines if a gantt chart is generated for run that
**                     shows each chunk's start/stop time. Defaults to 0 (No)
**
**  DIST_AUTOSIGNOFF   Determines if GridDistribute should sign off any idle
**                     host when no more chunks exist to schedule for work.
**                     NOTE: If this is set, RCollect will not work because
**                     the remote servers will have gone away at signoff.
**                     Defaults to 0 (no)
**
**  DIST_SOURCE        Determines if the source code will show up in the
**                     remote logs. Defaults to 0 (No)
**
**  DIST_DEBUG         Determines if debugging information will be generated.
**                     Defaults to 0 (No)
**
**  DIST_RETRYS        Determines number of retrys for failed signons and job
**                     execution. Defaults to 5.
**
**  DIST_DELAY         Number of seconds to sleep between checks of remote signon
**                     and remote execution status. Default is 2 seconds.
**
**  EmailTime          Number of minutes between email updates.
**                     If not defined or set to 0, no email is sent.
**                     If set to < 0, it will only send an email at the end.
**                     If set to > 0, it will send every X minutes and at end.
**
**  EmailTo            Address of email recipient. Required if sending email.
**
**                     If you are going to send email, you must make sure the
**                     following options are specified in the SAS config file
**                     or the SAS command line to enable email sending
**                        -EMAILHOST=
**                        -EMAILSYS=
**                        -EMAILPORT=
**                        -EMAILID=
**                        -EMAILPW=
**                        -EMAILAUTHPROTOCOL=
**
**  EmailSubject       Subject to put in email. If not provided, generic
**                     one will be provided. Date and time are appended to
**                     subject.
**
**-------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
**
**  GridInitParms Macro
**
**  This macro will initialize the GridDistribute system control parameters
**
**-------------------------------------------------------------------------*/
%macro GridInitParms;

    %global NIterAll;
    %global NIter;
    %global GridAppServer;
    %global MaxGridNodesToUse;
    %global MaxRunTime;
    %global EmailTime;
    %global EmailTo;
    %global EmailSubject;

    %global DistLog;
    %global DistLst;

    %if ("&DistLog" eq "") %then %let DistLog = distribute.log;
    %if ("&DistLst" eq "") %then %let DistLst = distribute.lst;


    %global DIST_STATUS;
    %global DIST_DETAIL;
    %global DIST_CHART;
    %global DIST_AUTOSIGNOFF;
    %global DIST_SOURCE;
    %global DIST_DEBUG;
    %global DIST_RETRYS;
    %global DIST_DELAY;

    %if ("&DIST_STATUS"       eq "") %then %let DIST_STATUS      = 1;
    %if ("&DIST_DETAIL"       eq "") %then %let DIST_DETAIL      = 1;
    %if ("&DIST_CHART"        eq "") %then %let DIST_CHART       = 0;
    %if ("&DIST_AUTOSIGNOFF"  eq "") %then %let DIST_AUTOSIGNOFF = 0;
    %if ("&DIST_SOURCE"       eq "") %then %let DIST_SOURCE      = 0;
    %if ("&DIST_DEBUG"        eq "") %then %let DIST_DEBUG       = 0;
    %if ("&DIST_RETRYS"       eq "") %then %let DIST_RETRYS      = 5;
    %if ("&DIST_DELAY"        eq "") %then %let DIST_DELAY       = 2;

    %global DIST_ClientRSub;
    %global DIST_Macros;

%mend GridInitParms;

%GridInitParms

/*---------------------------------------------------------------------------
**
**  GridInitHostVars Macro
**
**  This macro will initialize all variables used by a single grid host
**
**      Grid host state can be the following:
**       0 - awaiting signon submission
**       1 - signon occurred, awaiting completion
**       2 - signon completed, awaiting init rsubmit completion
**       3 - signon completed, awaiting job rsubmit completion
**       4 - signon completed, signoff completed
**       9 - signon retries failed - host eliminated
**
**      Signon states can be the following (per SAS/CONNECT):
**       0 - signon completed successfully
**       1 - signon failed
**       2 - already signed on to current session
**       3 - signon is in progress
**
**      Rsubmit states can be the following (per SAS/CONNECT):
**       0 - rsubmit completed successfully
**       1 - rsubmit failed
**       2 - rsubmit is in progress
**
**-------------------------------------------------------------------------*/
%macro GridInitHostVars(id);

    %global gridHostInitialized&id;
    %global gridHost&id;
    %global gridHostRetry&id;
    %global gridHostState&id;
    %global gridHostJob&id;
    %global gridSignonState&id;
    %global gridRsubmitState&id;

    %if (&&gridHostInitialized&id ne 1) %then %let gridHostInitialized&id = 0;

    %if (&&gridHostInitialized&id eq 0) %then %do;
        %if (&DIST_DEBUG eq 1) %then %put Initializing host &id;
        %let gridHostInitialized&id = 1;
        %let gridHost&id            = "";
        %let gridHostRetry&id       = &DIST_RETRYS;
        %let gridHostState&id       = 0;   /* State = awaiting submission */
        %let gridHostJob&id         = 0;
        %let gridSignonState&id     = 3;   /* State = In progress         */
        %let gridRsubmitState&id    = 3;   /* State = In progress         */
    %end;

%mend GridInitHostVars;

/*---------------------------------------------------------------------------
**
**  SignOn Macro.
**
**  This macro will initialize the system to start the signon process.
**
**-------------------------------------------------------------------------*/
%macro Signon;

    %global gridEnabled;        /* Is Grid Enabled flag                    */
    %global maxHosts;           /* Maximum number of hosts                 */
    %global badHosts;           /* Number of hosts where signon failed     */
    %global MaxGridNodesToUse;  /* Maximum number of grid nodes to use     */

    %local cnt;                 /* Loop counter                            */
    %local notesopt;            /* Saved notes options for DebugOutput*    */
    %local iHost;               /* Loop id of host we are evaluating       */

    /*---------------------------------------------------------------------
    ** Initialize the parameters just in case they have not been by now
    **-------------------------------------------------------------------*/
    %GridInitParms

    /*---------------------------------------------------------------------
    ** Turn on or off debugging info
    **-------------------------------------------------------------------*/
    %DebugOutputOn(&DIST_DEBUG);

    /*---------------------------------------------------------------------
    ** Check to see whether ClientRSub and Macros are defined.
    **-------------------------------------------------------------------*/
    proc catalog cat=work.sasmacr;
       contents out=_Macros;
    run;

    %let DIST_ClientRSub = 0;
    data _null_;
       set _Macros(where=(upcase(Name) = upcase("ClientRSub")));
       call symput('DIST_ClientRSub','1');
    run;

    %let DIST_Macros = 0;
    data _null_;
       set _Macros(where=(upcase(Name) = upcase("Macros")));
       call symput('DIST_Macros','1');
    run;

    /*---------------------------------------------------------------------
    ** Try to log onto the grid to see if we are grid enabled to determine
    ** the number of hosts we have to run jobs
    **-------------------------------------------------------------------*/
    %if ("&GridAppServer" ne "") %then %do;
        %let rc=%sysfunc(grdsvc_enable(_all_,"resource=&GridAppServer"));

        %if (&rc eq 0) %then %do;
            %put GridDistribute: Grid Enabled!;

            %let gridEnabled = 1;
            %let maxHosts    = %eval(%sysfunc(grdsvc_nnodes("resource=&GridAppServer"));
            %put GridDistribute: Maximum number of nodes on grid is &maxHosts;
        %end;
    %end;

    %else %do;
        %let rc = 1;
    %end;

    /*---------------------------------------------------------------------
    ** If there is no grid server available, set things up to spawn
    ** sessions locally setting hosts to the number of CPUs.
    **-------------------------------------------------------------------*/
    %if (&rc ne 0) %then %do;
        %put GridDistribute: Grid is NOT Enabled - spawning locally.;

        %let sascmdopt = %sysfunc(getoption(sascmd));
        %if ("&sascmdopt" eq "") %then options sascmd="!sascmd" %str(;);

        %let gridEnabled = 0;
        %let maxHosts    = %sysfunc(getoption(cpucount));

    %end;

    %put GridDistribute: Maximum number of nodes is &maxHosts;

    /*---------------------------------------------------------------------
    ** Make sure we are not using more than we are supposed to...
    **-------------------------------------------------------------------*/
    %if ("&MaxGridNodesToUse" eq "") %then %let MaxGridNodesToUse=0;
    %if ( &MaxGridNodesToUse  lt 0 ) %then %let MaxGridNodesToUse=0;

    %if ((&MaxGridNodesToUse gt 0) and (&MaxGridNodesToUse lt &maxHosts)) %then %do;
        %let maxHosts = &MaxGridNodesToUse;
        %put GridDistribute: Restricting number of hosts to &maxHosts;
    %end;

    /*---------------------------------------------------------------------
    ** Initialize the state variables for all grid hosts
    **-------------------------------------------------------------------*/
    %do cnt=1 %to &maxHosts;

        %GridInitHostVars(&cnt)

    %end;

    /*---------------------------------------------------------------------
    ** Do all other global initializations
    **-------------------------------------------------------------------*/
    %let badHosts = 0;

    /*---------------------------------------------------------------------
    ** Disable debugging
    **-------------------------------------------------------------------*/
    %DebugOutputOff(&DIST_DEBUG)

%mend Signon;

/*---------------------------------------------------------------------------
**
**  SignOff Macro.
**
**  This macro will actually signoff from all grid nodes.
**
**-------------------------------------------------------------------------*/
%macro Signoff;

    %local notesopt;            /* Saved notes options for DebugOutput*    */

    /*---------------------------------------------------------------------
    ** Turn on or off debugging info
    **-------------------------------------------------------------------*/
    %DebugOutputOn(&DIST_DEBUG)

    /*---------------------------------------------------------------------
    ** Signoff everyone
    **-------------------------------------------------------------------*/
    %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,0)

    signoff _all_;

    %RedirectOutputOff(&DIST_DEBUG)

    /*---------------------------------------------------------------------
    ** Disable debugging
    **-------------------------------------------------------------------*/
    %DebugOutputOff(&DIST_DEBUG)

    /*---------------------------------------------------------------------
    ** Initialize the state variables for all grid hosts to uninitialized
    **-------------------------------------------------------------------*/
    %do cnt=1 %to &maxHosts;

        %if (&DIST_DEBUG eq 1) %then %put Uninitializing host &cnt;
        %let gridHostInitialized&cnt=0;

    %end;

%mend Signoff;


/*---------------------------------------------------------------------------
**
**  Distribute Macro.
**
**  This macro will operate a remote signon/submission/signoff state machine
**
**-------------------------------------------------------------------------*/
%macro Distribute;

    %local cnt;                 /* Loop counter                            */
    %local lastEmailTime;       /* Last time email was sent                */
    %local gridElapsedTime;     /* Amount of time we have spent running    */
    %local notesopt;            /* Saved notes options                     */
    %local gridJobCheckStart;   /* Job # to use to start job complete check*/
    %local gridJobCount;        /* Number of jobs                          */
    %local gridJobsActive;      /* Number of jobs currently being processed*/
    %local gridJobsComplete;    /* Number of jobs completed                */
    %local iHost;               /* Loop id of host we are evaluating       */
    %local runTime;             /* Number of seconds allow to distribute   */
    %local linesizeOpt;

    /*---------------------------------------------------------------------
    ** Initialize the parameters just in case they have not bee by now
    **-------------------------------------------------------------------*/
    %GridInitParms;

    /*---------------------------------------------------------------------
    ** Save the linesize and set it to the max for printing status
    **-------------------------------------------------------------------*/
    %let linesizeOpt = %sysfunc(getoption(linesize));
    %if (&DIST_STATUS) %then
        options linesize=MAX;

    /*---------------------------------------------------------------------
    ** Turn on or off debugging info
    **-------------------------------------------------------------------*/
    %DebugOutputOn(&DIST_DEBUG)

    /*---------------------------------------------------------------------
    ** Set the sleep interval
    **-------------------------------------------------------------------*/
    %let gridNumSecs=&DIST_DELAY;
    %let gridUnits=1;

    /*---------------------------------------------------------------------
    ** Calculate the number of jobs we are going to have and reset the
    ** the number we have completed.
    **-------------------------------------------------------------------*/
    %if (("&NIterAll" eq "") or ("&NIter" eq "")) %then %do;

        %put Macro Variable NIter or NIterAll is not initialized;

        %let NIterAll = 0;
        %let NIter    = 1;
    %end;

    %let gridJobCount      = %eval((&NIterAll+&NIter-1)/&NIter);
    %let gridJobsComplete  = 0;
    %let gridJobsActive    = 0;
    %let gridJobCheckStart = 1;

    /*---------------------------------------------------------------------
    ** Do not use more grid nodes than we have jobs...
    **-------------------------------------------------------------------*/
    %if (&gridJobCount lt &maxHosts) %then %do;
        %let maxHosts = &gridJobCount;
        %put GridDistribute: Reducing number of hosts to number of jobs = &maxHosts;
    %end;

    /*---------------------------------------------------------------------
    ** Initialize the job state for all jobs to awaiting submission.
    **
    ** Grid job state can be the following:
    **  0 - awaiting submission
    **  1 - rsubmit occurred, awaiting rsubmit completion
    **  2 - rsubmit completed, done
    **  3 - retry count exceeded. Job cancelled
    **-------------------------------------------------------------------*/
    %do cnt=1 %to &gridJobCount;

        %let gridJobHost&cnt  = "";
        %let gridJobRetry&cnt = &DIST_RETRYS;
        %let gridJobState&cnt = 0; /* State = awaiting submission */

    %end;

    /*---------------------------------------------------------------------
    ** Save the start time, the email start time and convert the
    ** EmailTime to seconds from minutes.
    **-------------------------------------------------------------------*/
    %let gridTimeStart   = %sysfunc(datetime());
    %let gridElapsedTime = 0;
    %let lastEmailTime   = &gridTimeStart;

    %if ("&EmailTime" eq "") %then
        %let EmailTime = 0;
    %else
        %let EmailTime = %eval(&EmailTime * 60);

    /*---------------------------------------------------------------------
    ** Make sure we run the correct amount of time
    **-------------------------------------------------------------------*/
    %if ("&MaxRunTime" = "") %then %let runTime = 0;

    %if (&MaxRunTime <= 0) %then %do;
        %let runTime = 0;
    %end;

    %else %do;
        %let runTime = %eval(&MaxRunTime * 60);
        %put Run time limited to &runTime seconds;
    %end;

    /*---------------------------------------------------------------------
    ** Loop
    **  - While there are hosts                                  AND
    **  - While there are jobs to complete                       AND
    **  - While there is at least one host that has not gone bad AND
    **  - While the runtime has not expired or if it has, there
    **    are jobs still active.
    **-------------------------------------------------------------------*/
    %let iHost = 1;

    %do %while ((&maxHosts ne 0)                     and
                (&gridJobsComplete ne &gridJobCount) and
                (&badHosts ne &maxHosts)             and
                ((&runTime <= 0) or (&gridElapsedTime <= &runTime) or (&gridJobsActive > 0)));

        /*-----------------------------------------------------------------
        ** If the host state indicates it has not signed on and
        ** the maximum number of submitted jobs has not been reached,
        ** signon asynchronously to the grid and set the state to
        ** waiting for signon to complete.
        **---------------------------------------------------------------*/
        %if (&&gridHostState&iHost eq 0) %then %do;

            %if (&DIST_DEBUG) %then %put GridDistribute: Performing signon for host &iHost;

            %if (&&gridHostRetry&iHost ne 0) %then %do;

                %put GridDistribute: Signing on to Host #&iHost;

                %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,1)

                signon remote=Host&iHost signonwait=no cmacvar=gridSignonState&iHost;

                %RedirectOutputOff(&DIST_DEBUG)

                %let gridHostState&iHost = 1;
                %let gridHostRetry&iHost = %eval(&&gridHostRetry&iHost - 1);
                %let gridHostJobs&iHost  = 0;

            %end;

            %else %do;

                %put GridDistribute: Signon retry count for instance &iHost exceeded;

                %let gridHostState&iHost = 9;
                %let badHosts            = %eval(&badHosts + 1);

            %end;

        %end;

        /*-----------------------------------------------------------------
        ** If the host state indicates that the signon has been
        ** done, but has not completed, check to see if the signon has
        ** finished through the cmacvar value. If the signon has
        ** finished, initialize the host and update the host state to
        ** waiting for the init rsubmit to finish. If the signon failed,
        ** reset the state to try to signon again.
        **---------------------------------------------------------------*/
        %else %if (&&gridHostState&iHost eq 1) %then %do;

            /*-------------------------------------------------------------
            ** If the signon worked, submit the code for execution
            **-----------------------------------------------------------*/
            %if (&&gridSignonState&iHost eq 0) %then %do;

                %if (&DIST_DEBUG) %then
                    %if (&gridEnabled eq 1) %then
                        %put GridDistribute: Performing init rsubmit for host &iHost on host %sysfunc(grdsvc_getname(Host&iHost));
                    %else
                        %put GridDistribute: Performing init rsubmit for host &iHost on local machine;
                %else
                    %if (&gridEnabled eq 1) %then
                        %put GridDistribute: Host #&iHost is %sysfunc(grdsvc_getname(Host&iHost));
                    %else
                        %put GridDistribute: Host #&iHost is local machine;

                %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,0)

                /*---------------------------------------------------------
                ** Setup the RLS to get data sets from this host.
                **-------------------------------------------------------*/
                libname RWork&iHost slibref=WORK server=Host&iHost;

                /*---------------------------------------------------------
                ** Turn on or off source echo on the remote host.
                **-------------------------------------------------------*/
                %if (&DIST_SOURCE eq 0) %then %do;
                    rsubmit remote=Host&iHost wait=yes;
                        options nosource;
                    endrsubmit;
                %end;

                /*---------------------------------------------------------
                ** Instantiate the FirstRSub and TaskRSub macros on the
                ** remote host.
                **-------------------------------------------------------*/
                %RInit

                /*---------------------------------------------------------
                ** Instantiate any other client macros on the remote host.
                **-------------------------------------------------------*/
                %if (&DIST_Macros) %then %Macros;

                /*---------------------------------------------------------
                ** Instantiate the predefined per host macro variables on
                ** the remote host.
                **-------------------------------------------------------*/
                %if (&gridEnabled eq 1) %then
                    %let gridHost&iHost = %sysfunc(grdsvc_getname(Host&iHost));
                %else
                    %let gridHost&iHost = localHost;


                data _null_;
                    newseed = round(10000000*ranuni(&iHost));
                    call symput('seed', trim(left(put(newseed,best.))));
                run;


                %syslput Rem_Seed     = &seed            /remote=Host&iHost;
                %syslput Rem_Host     = &&gridHost&iHost /remote=Host&iHost;
                %syslput Rem_iHost    = &iHost           /remote=Host&iHost;
                %syslput Rem_NIterAll = &NIterAll        /remote=Host&iHost;
                %syslput Rem_NIter    = &NIter           /remote=Host&iHost;

                /*---------------------------------------------------------
                ** Submit the per host initialization code
                **-------------------------------------------------------*/
                rsubmit remote=Host&iHost connectwait=no cmacvar=gridRsubmitState&iHost;

                    %FirstRSub;

                endrsubmit;

                %RedirectOutputOff(&DIST_DEBUG)

                %if (&DIST_DEBUG) %then %put GridDistribute: Host &iHost initialization complete;

                %let gridHostState&iHost = 2;
            %end;

            /*-------------------------------------------------------------
            ** If the signon failed, set the host to retry job
            **-----------------------------------------------------------*/
            %else %if (&&gridSignonState&iHost eq 1) %then %do;

                %put GridDistribute: Signon failed for host &iHost - retrying;

                %let gridHostState&iHost = 0;
            %end;

        %end;

        /*-----------------------------------------------------------------
        ** If the host state indicates that the signon and init has
        ** been done, but has not completed, check to see if the init
        ** has finished through the cmacvar value. If the init has
        ** finished, find the next job to submit, setup the macros, and
        ** update the host state to waiting for the job rsubmit to
        ** finish. If the init failed, reset the state to try to init
        ** again.
        **---------------------------------------------------------------*/
        %else %if (&&gridHostState&iHost eq 2) %then %do;

            /*-------------------------------------------------------------
            ** Determine the elapsed time that Distribute has been running
            ** so we can limit the time if required. If runTime is not
            ** set, set the elapsed time to 0 so all jobs are submitted.
            **-----------------------------------------------------------*/
            %if (&runTime > 0) %then %do;
                %let now         = %sysfunc(datetime());
                %let elapsedTime = %sysevalf(&now - &gridTimeStart, integer);
            %end;

            %else %do;
                %let elapsedTime = -1;
            %end;

            /*-------------------------------------------------------------
            ** If the init rsubmit worked, set state to submit job.
            ** Set the job number to indicate we were initializing
            **-----------------------------------------------------------*/
            %if (&&gridRsubmitState&iHost eq 0) %then %do;

                /*---------------------------------------------------------
                ** Find the next job to process starting at the last
                ** finished job. We keep the starting index for runs that
                ** have a lot of jobs so that we do not spend all of our
                ** time looping through job macros.
                **-------------------------------------------------------*/
                %let jobNum = 0;

                %do job = &gridJobCheckStart %to &gridJobCount;

                    /*-----------------------------------------------------
                    ** If the job can be scheduled, check its retry count.
                    ** If the count has not been exceeded, schedule the
                    ** job. If it has been exceeded, set the job's state
                    ** to 'Retrys Exceeded'.
                    **---------------------------------------------------*/
                    %if (&&gridJobState&job eq 0) %then %do;

                        %if (&&gridJobRetry&job ne 0) %then %do;
                            %let jobNum = &job;
                            %let job    = &gridJobCount; /* Exit loop */
                        %end;

                        %else
                            %let gridJobState&jobNum = 3;

                    %end;
                %end;

                /*---------------------------------------------------------
                ** Set the starting job number for the next check to one
                ** more than the current job number or set it to one more
                ** than the total job count so that the loop will not
                ** execute the next time.
                **-------------------------------------------------------*/
                %if (&jobNum > 0) %then
                    %let gridJobCheckStart = %eval(&jobNum+1);
                %else
                    %let gridJobCheckStart = %eval(&gridJobCount+1);

                /*---------------------------------------------------------
                ** If a job is left to start...
                **-------------------------------------------------------*/
                %if ((&jobNum > 0) and (&elapsedTime <= &runTime)) %then %do;

                    /*-----------------------------------------------------
                    ** Save statistics information
                    **---------------------------------------------------*/
                    %let gridJobHost&jobNum       = &&gridHost&iHost;
                    %let gridJobSubmitTime&jobNum = %sysfunc(datetime());
                    %let gridJobRetry&jobNum      = %eval(&&gridJobRetry&jobNum - 1);
                    %let gridHostJob&iHost        = &jobNum;
                    %let numItersDone             = %eval((&gridJobsComplete * &NIter) + (&gridJobsActive * &NIter));

                    %let iters = %eval(&NIterALL - ((&jobNum-1) * &NIter));

                    %if (&iters >= &NIter) %then
                        %let gridJobIters&jobNum = &NIter;
                    %else
                        %let gridJobIters&jobNum = &iters;

                    %if (&DIST_DEBUG) %then %put GridDistribute: Job &JobNum started on host &&gridHost&iHost for &&gridJobIters&jobNum iterations;

                    /*-----------------------------------------------------
                    ** Instantiate the standard %Distribute macros on the
                    ** remote computer and save host name and iteration
                    ** count
                    **---------------------------------------------------*/
                    %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,0)

                    %syslput GlobalNSub   = &numItersDone          /remote=Host&iHost;
                    %syslput Rem_JobIters = &&gridJobIters&jobNum  /remote=Host&iHost;
                    %syslput Rem_JobID    = &jobNum                /remote=Host&iHost;

                    /*-----------------------------------------------------
                    ** Execute client per job macro statements
                    **---------------------------------------------------*/
                    %if (&DIST_ClientRSub) %then %ClientRSub;

                    /*-----------------------------------------------------
                    ** Submit the following code to be executed
                    **---------------------------------------------------*/
                    rsubmit remote=Host&iHost connectwait=no cmacvar=gridRsubmitState&iHost;

                        /*-------------------------------------------------
                        ** Record job start time
                        **-----------------------------------------------*/
                        data _null_;
                           now = datetime();
                           call symput('_gridJobStartTime',trim(left(put(now,best.))));
                        run;

                        /*-------------------------------------------------
                        ** Generate a new random number
                        **-----------------------------------------------*/
                        data _null_;
                            new_Seed = round(100000*ranuni(&Rem_Seed));
                            call symput('Rem_Seed',trim(left(put(new_Seed,best.))));
                        run;

                        /*-------------------------------------------------
                        ** Do main iterative task
                        **-----------------------------------------------*/
                        %TaskRSub;

                        /*-------------------------------------------------
                        ** Record time job ended
                        **-----------------------------------------------*/
                        data _null_;
                           now = datetime();
                           call symput('_gridJobEndTime',trim(left(put(now,best.))));
                        run;

                    endrsubmit;

                    %RedirectOutputOff(&DIST_DEBUG)

                    /*-----------------------------------------------------
                    ** Update the job and host states and active job cnt
                    **---------------------------------------------------*/
                    %let gridHostState&iHost = 3;
                    %let gridJobState&jobNum = 1;
                    %let gridJobsActive      = %eval(&gridJobsActive + 1);

                %end;

                /*---------------------------------------------------------
                ** If no job is available, signoff this machine to free
                ** it for other grid jobs.
                **-------------------------------------------------------*/
                %else %if (&DIST_AUTOSIGNOFF) %then %do;

                    %let gridHostState&iHost = 4;

                    %if (&gridEnabled eq 1) %then
                        %put GridDistribute: Performing signoff for host &iHost on host %sysfunc(grdsvc_getname(Host&iHost));
                    %else
                        %put GridDistribute: Performing signoff for host &iHost on local machine;

                    %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,1)

                    signoff remote=Host&iHost;

                    %RedirectOutputOff(&DIST_DEBUG)

                %end;

            %end;

            /*-------------------------------------------------------------
            ** If the init failed, set the host to retry job
            **-----------------------------------------------------------*/
            %else %if (&&gridSignonState&iHost eq 1) %then %do;

                %put GridDistribute: Init failed for host &iHost - retrying;

                %let gridHostState&iHost = 1;
            %end;

        %end;

        /*-----------------------------------------------------------------
        ** If the host state indicates that the signon and rsubmit
        ** has occurred, check the rsubmit cmacvar to see if the code
        ** has completed. If it has, set the state indicating
        ** that this job has completed, and increment the
        ** completed processing count.
        **---------------------------------------------------------------*/
        %else %if (&&gridHostState&iHost eq 3) %then %do;

            %if (&&gridRsubmitState&iHost ne 2) %then %do;

                /*---------------------------------------------------------
                ** Get the job number executing on this host
                **-------------------------------------------------------*/
                %let jobNum = &&gridHostJob&iHost;

                /*---------------------------------------------------------
                ** If the rsubmit was successful, set the job as
                ** complete, increment the complete count, and
                ** gather statistics. Also reset the host state so that
                ** it can accept another job.
                **-------------------------------------------------------*/
                %if (&&gridRsubmitState&iHost eq 0) %then %do;

                    %let gridJobState&jobNum = 2;
                    %let gridJobsComplete    = %eval(&gridJobsComplete + 1);

                    %if (&DIST_DEBUG) %then %put GridDistribute: Job &jobNum finished on &&gridHost&iHost, &gridJobsComplete jobs of &gridJobCount complete.;

                    %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,0)

                    rsubmit remote=Host&iHost wait=yes;

                        %nrstr(%%)sysrput gridJobStartTime&rem_JobID = &_gridJobStartTime;
                        %nrstr(%%)sysrput gridJobEndTime&rem_JobID   = &_gridJobEndTime;

                    endrsubmit;

                    %RedirectOutputOff(&DIST_DEBUG)

                    %let gridJobDoneTime&jobNum = %sysfunc(datetime());

                    /*-----------------------------------------------------
                    ** Set host & rsubmit state to enable to next job.
                    **---------------------------------------------------*/
                    %let gridHostState&iHost    = 2;
                    %let gridRsubmitState&iHost = 0;
                %end;

                /*---------------------------------------------------------
                ** If the rsubmit failed, set the job to be retried after
                ** signing off the host. Set the host to be tried to sign
                ** back on.
                **-------------------------------------------------------*/
                %else %do;

                    %put GridDistribute: Rsubmit failed for job &jobNum on host &iHost - retrying job;

                    /*-----------------------------------------------------
                    ** Signoff failed host
                    **---------------------------------------------------*/
                    %put GridDistribute: Performing signoff for host &iHost to enable re-signon;

                    %RedirectOutputOn(&DIST_DEBUG,&DistLog,&DistLst,1)

                    signoff remote=Host&iHost;

                    %RedirectOutputOff(&DIST_DEBUG)

                    /*-----------------------------------------------------
                    ** Set job, host, & rsubmit state to enable re-signing
                    ** on and the job can be retried.
                    **---------------------------------------------------*/
                    %let gridJobState&jobNum    = 0;
                    %let gridHostState&iHost    = 0;
                    %let gridRsubmitState&iHost = 0;

                    /*-----------------------------------------------------
                    ** If the next job to start the state check is greater
                    ** than this job number, reset the starting job number
                    ** for the next check to this job number.
                    **---------------------------------------------------*/
                    %if (&gridJobCheckStart gt &jobNum) %then
                        %let gridJobCheckStart = &jobNum;
                %end;

                /*---------------------------------------------------------
                ** Decrease the number of running jobs.
                **-------------------------------------------------------*/
                %let gridJobsActive = %eval(&gridJobsActive - 1);

            %end;

        %end;

        /*-----------------------------------------------------------------
        ** Increment the host ID.
        **---------------------------------------------------------------*/
        %let iHost = %eval(&iHost + 1);

        /*-----------------------------------------------------------------
        ** If we have gone through all hosts once...
        **---------------------------------------------------------------*/
        %if (&iHost > &maxHosts) %then %do;

            /*-------------------------------------------------------------
            ** Print out the status and determine various counts
            **-----------------------------------------------------------*/
            %let StatLine        = ;
            %let hostsDead       = 0;
            %let hostsIdle       = 0;
            %let hostsBusy       = 0;
            %let hostsSignedOff  = 0;
            %let currentTime     = %sysfunc(dateTime());
            %let gridElapsedTime = %sysevalf(&currentTime-&gridTimeStart,integer);

            data _null_;
                call symput('prettyTime', trim(left(put(&gridElapsedTime,time.))));
            run;

            %do cnt=1 %to &maxHosts;

                /*---------------------------------------------------------
                ** If we have not finished signing on, status = '?'
                **-------------------------------------------------------*/
                %if (&&gridHostState&cnt lt 2) %then %do;
                    %let StatLine  = &StatLine.?;
                    %let hostsBusy = %eval(&hostsBusy + 1);
                    %end;

                /*---------------------------------------------------------
                ** If the machine had an error, status = 'X'
                **-------------------------------------------------------*/
                %else %if (&&gridHostState&cnt eq 9) %then %do;
                    %let StatLine  = &StatLine.X;
                    %let hostsDead = %eval(&hostsDead + 1);
                    %end;

                /*---------------------------------------------------------
                ** If the machine was released, status = '-'
                **-------------------------------------------------------*/
                %else %if (&&gridHostState&cnt eq 4) %then %do;
                    %let StatLine       = &StatLine.-;
                    %let hostsSignedOff = %eval(&hostsSignedOff + 1);
                    %end;

                /*---------------------------------------------------------
                ** If the machine is busy, status = '.'
                **-------------------------------------------------------*/
                %else %if (&&gridRsubmitState&cnt eq 2) %then %do;
                    %let StatLine  = &StatLine..;
                    %let hostsBusy = %eval(&hostsBusy + 1);
                    %end;

                /*---------------------------------------------------------
                ** If the machine is idle, status = '!'
                **-------------------------------------------------------*/
                %else %do;
                    %let StatLine  = &StatLine.!;
                    %let hostsIdle = %eval(&hostsIdle + 1);
                    %end;

                /*---------------------------------------------------------
                ** See if we need to generate a new line
                **-------------------------------------------------------*/
                %let newline=%eval(&cnt - ((&cnt/100) * 100));

                %if ((&newline eq 0) and (&DIST_STATUS)) %then %do;
                    %put Stat: [&prettyTime] &Statline;
                    %let Statline = ;
                %end;
            %end;

            %let itersDone       = %eval(&gridJobsComplete * &NIter);
            %let itersSubmitted  = %eval((&gridJobsActive * &NIter) + &itersDone);

            %if (&itersSubmitted > &NIterAll) %then %let itersSubmitted=&NIterAll;

            %let StatLine = &StatLine (&itersSubmitted/&itersDone) &NIterAll;

            %if (&DIST_STATUS) %then %put Stat: [&prettyTime] &StatLine;

            /*---------------------------------------------------------------------
            ** Determine the elapsed time between emails and send an email if
            ** time has elapsed. If the number of jobs complete is equal to the
            ** number of jobs to run, skip the email since it will be sent at
            ** the end anyway.
            **-------------------------------------------------------------------*/
            %if (&EmailTime > 0) %then %do;

                %let emailElapsedTime = %sysevalf(&currentTime-&lastEmailTime,integer);

                %if ((&emailElapsedTime >= &EmailTime) and (&gridJobsComplete ne &gridJobCount)) %then %do;

                    %let lastEmailTime = %sysfunc(datetime());
                    %EmailStatus(&itersSubmitted, &itersDone, &gridElapsedTime, &currentTime)

                %end;

            %end;

            /*-------------------------------------------------------------
            ** Reset the host id back to the beginning
            **-----------------------------------------------------------*/
            %let iHost = 1;

            /*-------------------------------------------------------------
            ** If there are no idle hosts or there are no more iterations
            ** to submit, wait for specified number of seconds.
            **   (Yes, everything needs to be in macro variables...)
            **-----------------------------------------------------------*/
            %if ((&hostsIdle eq 0) or (&itersSubmitted ge &NIterAll)) %then %do;
                %syscall sleep(gridNumSecs,gridUnits);
            %end;

            /*-------------------------------------------------------------
            ** Check the number of grid nodes again to set max jobs since
            ** nodes may come and go if this job is long running. If new
            ** nodes are added, instatiate the host macro variables and
            ** update the max hosts count. We do not care if nodes get
            ** removed from the grid - they will generate errors.
            **-----------------------------------------------------------*/
            %if ((&gridEnabled eq 1)          and
                 (&gridJobCount gt &maxHosts) and
                 ((&MaxGridNodesToUse eq 0) or (&maxHosts lt &MaxGridNodesToUse))) %then %do;

                %let newMaxHosts = %eval(%sysfunc(grdsvc_nnodes("resource=&GridAppServer"));

                /*---------------------------------------------------------
                ** Make sure we are not using more nodes than we are
                ** supposed to.
                **-------------------------------------------------------*/
                %if ((&MaxGridNodesToUse gt 0) and (&MaxGridNodesToUse lt &newMaxHosts)) %then %do;
                    %let newMaxHosts = &MaxGridNodesToUse;
                %end;

                /*---------------------------------------------------------
                ** Update all control variables
                **-------------------------------------------------------*/
                %if &newMaxHosts > &maxHosts %then %do;

                    %do cnt=%eval(&maxHosts+1) %to &newMaxHosts;

                        %GridInitHostVars(&cnt)

                    %end;

                    %let maxHosts  = &newMaxHosts;
                    %let hostsIdle = %eval(&hostsIdle + (&newMaxHosts - &maxHosts));
                    %put GridDistribute: maxHosts has changed to &maxHosts;
                %end;

            %end;

            /*-------------------------------------------------------------
            ** In rare cases, we may have signed off of machines and then
            ** have a bad grid run requiring us to restart the grid job
            ** on a previously closed machine. If that is the case, try
            ** to re-signon to a machine in the grid to make a machine
            ** available for running the job.
            **-----------------------------------------------------------*/
            %let jobsAssigned = %eval(&gridJobsComplete + &gridJobsActive);
            %if ((&jobsAssigned ne &gridJobCount) and (&hostsIdle eq 0) and (&hostsSignedOff ne 0)) %then %do;

                %do cnt=1 %to &maxHosts;

                    %if (&&gridHostState&cnt eq 4) %then
                        %let gridHostState&cnt = 0;

                %end;

            %end;

        %end;

    %end;

    /*---------------------------------------------------------------------
    ** Determine the elapsed time
    **-------------------------------------------------------------------*/
    %let gridTimeEnd     = %sysfunc(datetime());
    %let gridElapsedTime = %sysevalf(&gridTimeEnd-&gridTimeStart,integer);

    %if ((&runTime > 0) and (&gridElapsedTime > &runTime)) %then
        %put GridDistribute: Run time exceeded max=&runTime, actual=&gridElapsedTime;

    /*---------------------------------------------------------------------
    ** Update the user via email if necessary
    **-------------------------------------------------------------------*/
    %if (&EmailTime ge 0) %then %do;

        %let itersDone = %eval(&gridJobsComplete * &NIter);

        %EmailStatus(&itersDone, &itersDone, &gridElapsedTime, &gridTimeEnd)

    %end;

    /*---------------------------------------------------------------------
    ** Restore the linesize
    **-------------------------------------------------------------------*/
    options linesize=&linesizeOpt;

    /*---------------------------------------------------------------------
    ** Print out statistics if we had hosts to execute on...
    **-------------------------------------------------------------------*/
    %if ((&maxHosts ne 0) and (&badHosts ne &maxHosts)) %then %do;

        /*-----------------------------------------------------------------
        ** Count the number of distinct host names we actually ran on
        **---------------------------------------------------------------*/
        data HostNames;

           do job=1 to &gridJobsComplete;
               Host = symget('gridJobHost' || trim(left(put(job,best.))));
               output;
           end;
           stop;

        run;

        proc sql nonumber noprint;
            select count(DISTINCT Host) into :NumHosts from HostNames;
        run;

        /*-----------------------------------------------------------------
        ** Calculate the statistics (the input() function is to prevent
        ** the note that says that it converted character data to numeric)
        **---------------------------------------------------------------*/
        data TimeSumm;
           keep jobNum Host NIter TimeSubmit TimeDone TimeTotal TimeJobStart TimeJobEnd TimeJobTotal;
           format TimeSubmit tod. TimeDone tod. TimeTotal time. TimeJobStart tod. TimeJobEnd tod. TimeJobTotal time.;

           do jobNum = 1 to &gridJobsComplete;

               Host         =       symget('gridJobHost'       || trim(left(put(jobNum,best.))));        *put Host;
               NIter        =       symget('gridJobIters'      || trim(left(put(jobNum,best.))));        *put NIter;

               TimeSubmit   = input(symget('gridJobSubmitTime' || trim(left(put(jobNum,best.)))),best.); *put TimeSubmit;
               TimeDone     = input(symget('gridJobDoneTime'   || trim(left(put(jobNum,best.)))),best.); *put TimeDone;

               TimeJobStart = input(symget('gridJobStartTime'  || trim(left(put(jobNum,best.)))),best.); *put TimeJobStart;
               TimeJobEnd   = input(symget('gridJobEndTime'    || trim(left(put(jobNum,best.)))),best.); *put TimeJobEnd;

               TimeTotal    = TimeDone - TimeSubmit;                                                     *put TimeTotal;
               TimeJobTotal = TimeJobEnd - TimeJobStart;                                                 *put TimeJobTotal;

               output;

           end;
           stop;
        run;

        /*-----------------------------------------------------------------
        ** Create the gantt chart if requested
        **---------------------------------------------------------------*/
        %if (&DIST_CHART ne 0) %then %do;

            proc gantt data=TimeSumm;
              chart /BS=TimeSubmit
                     BF=TimeDone
                     AS=TimeJobStart
                     AF=TimeJobEnd
                     mindate=&gridTimeStart
                     maxdate=&gridTimeEnd
                     nolegend
                     fill
                     dupok
                     pcompress
                     useformat;
              id Host;
            run;

            proc sort data=TimeSumm out=TimeSumm;
                by Host TimeSubmit;
            run;

            proc gantt data=TimeSumm;
              by Host;
              chart /BS=TimeSubmit
                     BF=TimeDone
                     AS=TimeJobStart
                     AF=TimeJobEnd
                     mindate=&gridTimeStart
                     maxdate=&gridTimeEnd
                     nolegend
                     fill
                     pcompress
                     useformat;
            run;

        %end;

        /*-----------------------------------------------------------------
        ** Output the detailed info if requested
        **---------------------------------------------------------------*/
        %if (&DIST_DETAIL ne 0) %then %do;

            proc sort data=TimeSumm out=TimeSumm;
                by TimeSubmit;
            run;

            proc print data=TimeSumm label noobs;
                label  jobNum          = "Job #";
                label  NIter           = "No. Iter";
                label  TimeSubmit      = "Submit Time";
                label  TimeDone        = "End Time";
                label  TimeTotal       = "Total Run Time";
                label  TimeJobStart    = "Node Start Time";
                label  TimeJobEnd      = "Node End Time";
                label  TimeJobTotal    = "Total Run Time on Node";
                id jobNum Host NIter TimeSubmit TimeDone TimeTotal TimeJobStart TimeJobEnd TimeJobTotal;
            run;

        %end;


        /*-----------------------------------------------------------------
        ** Summarize data
        **---------------------------------------------------------------*/
        proc summary data=TimeSumm print;
          var TimeTotal TimeJobTotal;
          output out=TotalTime Sum=TimeTotal TimeJobTotal;
        run;

        /*-----------------------------------------------------------------
        ** Output the total statistics to the log
        **---------------------------------------------------------------*/
        data _null_;

            set TotalTime;

            StartTime       = &gridTimeStart;
            EndTime         = &gridTimeEnd;
            Elapsed         = &gridElapsedTime;
            PerJobTotal     = TimeTotal     /&gridJobsComplete;
            PerJobJobTotal  = TimeJobTotal  /&gridJobsComplete;
            PerHostTotal    = TimeTotal     /&NumHosts;
            PerHostJobTotal = TimeJobTotal  /&NumHosts;
            PercentTotal    = ((TimeTotal-Elapsed)   /&NumHosts)/Elapsed;
            PercentJobTotal = ((TimeJobTotal-Elapsed)/&NumHosts)/Elapsed;

            put " ";
            put "               Start time:                    " StartTime       tod12.;
            put "               End Time:                      " EndTime         tod12.;
            put "               Total elapsed time:            " Elapsed         time12.;
            put "               Sum total time running:        " TimeTotal       time12.;
            put "               Sum total time in job:         " TimeJobTotal    time12.;
            put "               Per job time running:          " PerJobTotal     time12.;
            put "               Per job time in job:           " PerJobJobTotal  time12.;
            put "               Per host time running:         " PerHostTotal    time12.;
            put "               Per host time in job:          " PerHostJobTotal time12.;
            put "               Percent speed increase:        " PercentTotal    percent12.;
            put "               Percent speed increase in job: " PercentJobTotal percent12.;
            put " ";
            put " ";
        run;

    %end;

    /*---------------------------------------------------------------------
    ** Disable debugging
    **-------------------------------------------------------------------*/
    %DebugOutputOff(&DIST_DEBUG)


%mend Distribute;

/*---------------------------------------------------------------------------
**
**  MACRO:       RCollect
**
**  USAGE:       %RCollect(<<Remote DS prefix>>,<<Client DS>>);
**
**  DESCRIPTION: This macro collects similarly named data sets from &n
**               libraries named RWork1, ..., RWork&n, into a single data set.
**
**  NOTES:       In order to use a view to virtually collect the data sets,
**               use 'dsname / view=dsname' as the client data set name.
**
**-------------------------------------------------------------------------*/
%macro RCollect(from, to, n=, by=);

    %global maxHosts;

    %if (^%length(%left(%trim(&n)))) %then %let n=&maxHosts;

    %do iHost=1 %to &maxHosts;
        %global Host&iHost gridHostState&iHost;
    %end;

    data &to;

       %if (^%length(%left(%trim(&by)))) %then %do;
          set
       %end;
       %else %do;
          merge
       %end;

       %do i=1 %to &N;
           %if (&&gridHostState&i = 2) %then %do;
               RWork&i..&from
           %end;
       %end;

       ;

       %if (%length(%left(%trim(&by)))) %then %do;
           by &by;
       %end;
    run;

%mend RCollect;

/*---------------------------------------------------------------------------
**
** EmailStatus
**
** Internal macro that will email the user with a status update.
**
**-------------------------------------------------------------------------*/
%macro EmailStatus(itersSubmitted, itersDone, elapsedTime, currentTime);

    %local notesopt; %let notesopt=%sysfunc(getoption(notes)); options nonotes;

    %let emailHost = %sysfunc(getoption(emailhost));
    %let emailSys  = %sysfunc(getoption(emailsys));

    %if (("&EmailTo" ne "") and ("&emailHost" ne "") and ("&emailSys" ne "")) %then %do;

        /*-------------------------------------------------------------------
        ** Create the subject line
        **-----------------------------------------------------------------*/
        data _null_;
            call symput('currentTime',trim(left(put(&currentTime,datetime18.2))));
        run;

        %if ("&EmailSubject" eq "") %then
            %let EmailSubject=GridDistribute Update;

        %let subj=&EmailSubject at &currentTime;

        %if (&itersSubmitted eq &itersDone) %then
            %let subj=COMPLETE: &subj;

        /*-------------------------------------------------------------------
        ** Send the email
        **-----------------------------------------------------------------*/
        filename outbox email "&EmailTo" subject="&subj";

        data _null_;
           file outbox;

           itersSubmitted = trim(left(put(&itersSubmitted,best.)));
           itersDone      = trim(left(put(&itersDone,best.)));
           elapsedTime    = trim(left(put(&elapsedTime,time11.2)));

           put 'Elapsed Time         = ' elapsedTime;
           put 'Iterations Submitted = ' itersSubmitted;
           put 'Iterations Done      = ' itersDone;
        run;

    %end;

    options &notesopt;

%mend;

/*---------------------------------------------------------------------------
** Debug and Output Redirection Macros
**-------------------------------------------------------------------------*/
%macro DebugOutputOn(DebugFlag,LogFile,ListFile);

    %if (&DebugFlag eq 0) %then %do;
        %let notesopt = %sysfunc(getoption(notes));
        options nonotes;
    %end;

    %else %do;
        options mprint mtrace;
    %end;

%mend;

%macro DebugOutputOff(DebugFlag);

    %if (&DebugFlag eq 0) %then %do;
        options &notesopt;
    %end;

%mend;

%macro RedirectOutputOn(DebugFlag,LogFile,ListFile,AllowNotes);

    %if (&DebugFlag eq 0) %then %do;
        %if (&AllowNotes ne 0) %then %do;
            options notes;
        %end;
        proc printto log="&LogFile" print="&ListFile"; run;
    %end;

%mend;

%macro RedirectOutputOff(DebugFlag);

    %if (&DebugFlag eq 0) %then %do;
        %let stimeropt = %sysfunc(getoption(stimer));
        options nostimer;
        proc printto; run;
        options &stimeropt;
        options nonotes;
    %end;

%mend;

